import clickhouse_connect
import duckdb
import numpy as np
import os
import pandas as pd
import sqlalchemy
from sqlalchemy.orm import Session
import time
from typing import List
from zaturn.tools import config


def list_tables(source):
    try:
        match source['source_type']:
            case "sqlite":
                result = execute_query(source,
                    "SELECT name FROM sqlite_schema WHERE type ='table' AND name NOT LIKE 'sqlite_%';"
                )
                return result['name'].to_list()

            case "postgresql":
                result = execute_query(source,
                    "SELECT tablename FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog' AND schemaname != 'information_schema';"
                )
                return result['tablename'].to_list()

            case "mysql":
                result = execute_query(source, "SHOW TABLES")
                for col in list(result):
                    if col.startswith("Tables_in_"):
                        return result[col].to_list()

            case "mssql":
                result = execute_query(source, "SELECT name FROM sys.tables")
                return result['name'].to_list()
                
            case "duckdb" | "csv" | "parquet" | "clickhouse":
                result = execute_query(source, "SHOW TABLES")
                return result['name'].to_list()

    except Exception as e:
        return str(e)


def describe_table(source, table_name):
    match source['source_type']:
        case 'sqlite':
            return execute_query(source,
                f'PRAGMA table_info("{table_name}");'
            )
            
        case 'postgresql' | 'mssql':
            return execute_query(source,
                f"SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE table_name = '{table_name}';"
            )
            
        case "mysql" | "duckdb" | "csv" | "parquet" | "clickhouse":
            if ' ' in table_name:
                table_name = f'`{table_name}`'
                
            return execute_query(source,
                f'DESCRIBE {table_name};'
            )
            

def execute_query(source: dict, query: str):
    """Run the query using the appropriate engine and read only config"""
    url = source['url']
                
    match source['source_type']:
        case "sqlite":
            with sqlalchemy.create_engine(url).connect() as conn:
                conn.execute(sqlalchemy.text('PRAGMA query_only = ON;'))
                result = conn.execute(sqlalchemy.text(query))
                return pd.DataFrame(result)

        case "mysql":
            engine = sqlalchemy.create_engine(url)
            with Session(engine) as session:
                session.autoflush = False
                session.autocommit = False
                session.flush = lambda *args: None
                session.execute(sqlalchemy.text('SET SESSION TRANSACTION READ ONLY;'))
                result = session.execute(sqlalchemy.text(query))
                return pd.DataFrame(result)

        case "mssql":
            engine = sqlalchemy.create_engine(url)
            with Session(engine) as session:
                # no known way to ensure read-only here
                # please use read-only credentials
                result = session.execute(sqlalchemy.text(query))
                return pd.DataFrame(result)

        case "postgresql":
            engine = sqlalchemy.create_engine(url)
            with engine.connect() as conn:
                conn = conn.execution_options(
                    isolation_level="SERIALIZABLE",
                    postgresql_readonly=True,
                    postgresql_deferrable=True,
                )
                with conn.begin():
                    result = conn.execute(sqlalchemy.text(query))
                    return pd.DataFrame(result)

        case "clickhouse":
            client = clickhouse_connect.get_client(dsn=url)
            client.query('SET readonly=1;')
            return client.query_df(query, use_extended_dtypes=False)

        case "duckdb":
            conn = duckdb.connect(url, read_only=True)
            return conn.execute(query).df()

        case "csv":
            conn = duckdb.connect(database=':memory:')
            conn.execute(f"CREATE VIEW CSV AS SELECT * FROM read_csv('{url}')")
            return conn.execute(query).df()

        case "parquet":
            conn = duckdb.connect(database=':memory:')
            conn.execute(f"CREATE VIEW PARQUET AS SELECT * FROM read_parquet('{url}')")
            return conn.execute(query).df()
        
        case _:
            raise Exception("Unsupported Source")


def save_query(df: pd.DataFrame):
    """Save query results to disk and return a unique reference id"""
    query_id = 'q' + str(int(time.time()))
    filepath = os.path.join(config.QUERIES_DIR, f'{query_id}.parquet')
    df.replace({np.nan: None}).to_parquet(filepath, engine='pyarrow', index=False)
    return query_id


def load_query(query_id: str):
    """Load query results from disk using unique reference id"""
    filepath = os.path.join(config.QUERIES_DIR, f'{query_id}.parquet')
    df = pd.read_parquet(filepath, engine='pyarrow').replace({np.nan: None})
    df.reset_index(drop=True, inplace=True)
    return df



from datetime import datetime
import json
import os
from pathlib import Path
import urllib.parse


import platformdirs
from slugify import slugify


USER_DATA_DIR = Path(platformdirs.user_data_dir('zaturn', 'zaturn'))
STATE_FILE = USER_DATA_DIR / 'studio.json'
CHATS_DIR = USER_DATA_DIR / 'chats'
os.makedirs(CHATS_DIR, exist_ok=True)


def load_state() -> dict:
    if os.path.exists(STATE_FILE):
        with open(STATE_FILE) as f:
            state = json.loads(f.read())
            state['sources'] = state.get('sources', {})
            return state
    else:
        return {}


def save_state(state: dict):
    with open(STATE_FILE, 'w') as f:
        f.write(json.dumps(state, indent=2))


def save_datafile(datafile, filename: str) -> Path:
    target_dir = Path(USER_DATA_DIR) / 'studio_data'
    os.makedirs(target_dir, exist_ok=True)

    target_path = target_dir / filename
    datafile.save(target_path)
    return target_path


def remove_datafile(filepath):
    if filepath.startswith("sqlite:///"):
        filepath = filepath.replace("sqlite:///", "")
    os.remove(filepath)


def create_chat(question: str):
    slug = slugify(question[:20]).strip("-")
    slug += '-' + str(hex(int(datetime.now().timestamp() * 1000000)))[2:]
    
    chat = {
        'slug': slug,
        'messages': [{
            'role': 'user',
            'content': question,
        }]
    }
    
    filename = CHATS_DIR / f'{slug}.json'
    with open(filename, 'w') as f:
        f.write(json.dumps(chat, indent=2))
        
    return slug


def load_chat(slug: str):
    try:
        with open(CHATS_DIR / f'{slug}.json') as f:
            return json.loads(f.read())
    except:
        return None


def save_chat(slug: str, chat: dict):
    filename = CHATS_DIR / f'{slug}.json'
    with open(filename, 'w') as f:
        f.write(json.dumps(chat, indent=2))


def list_chats():
    return sorted(
        [Path(f).stem for f in os.listdir(CHATS_DIR)],
        key = lambda stem: os.path.getctime(CHATS_DIR / f'{stem}.json'),
    )[::-1]



from typing import Any, List, Union, Annotated

from pydantic import Field

from zaturn.tools import query_utils


class Core:

    def __init__(self, data_sources): 
        self.data_sources = data_sources
        self.tools = [
            self.list_data_sources,
            self.describe_table,
            self.run_query,
        ]


    def list_data_sources(self) -> str:
        """
        List all available data sources.
        Returns a list of unique source_ids to be used for other queries.
        Source type is included in the source_id string.
        While drafting SQL queries use appropriate syntax as per source type.
        """
        try:
            if not self.data_sources:
                return "No data sources available. Add data sources."
            
            result = "Available data sources:\n\n"
            for source_id in self.data_sources:
                tables = query_utils.list_tables(
                    self.data_sources[source_id]
                )
                if type(tables) is list:
                    tables = ', '.join(tables)
                result += f"- {source_id}\nHas tables: {tables}\n"
                
            return result
            
        except Exception as e:
            return str(e)


    def describe_table(self, 
        source_id: Annotated[
            str, Field(description='The data source')
        ], 
        table_name: Annotated[
            str, Field(description='The table in the data source')
        ]
    ) -> str:
        """
        Lists columns and their types in the specified table of specified data source.
        """
        
        try:
            source = self.data_sources.get(source_id)
            if not source:
                return f"Source {source_id} Not Found"
    
            result = query_utils.describe_table(source, table_name)
            return result.to_markdown(index=False)
        
        except Exception as e:
            return str(e)


    def run_query(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ]
    ) -> str:
        """
        Run query against specified source
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name for csv sources.
        Use 'PARQUET' as the table name for parquet sources.
    
        This will return a dataframe with the results.
        """
        
        try:
            source = self.data_sources.get(source_id)
            if not source:
                return f"Source {source_id} Not Found"
                
            df = query_utils.execute_query(source, query)
            return df.to_markdown(index=False)
        except Exception as e:
            return str(e)






from zaturn.tools import core, visualizations


class ZaturnTools:

    def __init__(self, data_sources):
        self.tools = [
            *core.Core(data_sources).tools,
            *visualizations.Visualizations(data_sources).tools,
        ]
        

    



from base64 import b64encode
from typing import Any, List, Union, Annotated

from mcp.types import ImageContent
import plotly.express as px
from pydantic import Field

from zaturn.tools import query_utils




def _fig_to_image(fig):
    fig_encoded = b64encode(fig.to_image(format='png')).decode()
    img_b64 = "data:image/png;base64," + fig_encoded
    
    return ImageContent(
        type = 'image',
        data = fig_encoded,
        mimeType = 'image/png',
        annotations = None,
    )


class Visualizations:

    def __init__(self, data_sources): 
        self.data_sources = data_sources
        self.tools = [
            self.scatter_plot,
            self.line_plot,
            self.histogram,
            self.strip_plot,
            self.box_plot,
            self.bar_plot,
        ]


    def _get_df_from_source(self, source_id, query):
        source = self.data_sources.get(source_id)
        if not source:
            raise Exception(f"Source {source_id} Not Found")
                
        return query_utils.execute_query(source, query)
            

    def scatter_plot(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ],
        x: Annotated[
            str, Field(description='Column name from SQL result to use for x-axis')
        ],
        y: Annotated[
            str, Field(description='Column name from SQL result to use for y-axis')
        ],
        color: Annotated[
            str | None, Field(description='Optional; column name from SQL result to use for coloring the points, with color representing another dimension')
        ] = None,
    ) -> str:
        """
        Run query against specified source and make a scatter plot using result
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name in the SQL query for csv sources.
        Use 'PARQUET' as the table name in the SQL query for parquet sources.
    
        This will return an image of the plot
        """

        try:
            df = self._get_df_from_source(source_id, query)
            fig = px.scatter(df, x=x, y=y, color=color)
            fig.update_xaxes(autotickangles=[0, 45, 60, 90])

            return _fig_to_image(fig)
        except Exception as e:
            return str(e)


    def line_plot(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ],
        x: Annotated[
            str, Field(description='Column name from SQL result to use for x-axis')
        ],
        y: Annotated[
            str, Field(description='Column name from SQL result to use for y-axis')
        ],
        color: Annotated[
            str | None, Field(description='Optional; column name from SQL result to use for drawing multiple colored lines representing another dimension')
        ] = None,
    ) -> str:
        """
        Run query against specified source and make a line plot using result
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name in the SQL query for csv sources.
        Use 'PARQUET' as the table name in the SQL query for parquet sources.
    
        This will return an image of the plot
        """

        try:
            df = self._get_df_from_source(source_id, query)
            fig = px.line(df, x=x, y=y, color=color)
            fig.update_xaxes(autotickangles=[0, 45, 60, 90])

            return _fig_to_image(fig)
        except Exception as e:
            return str(e)


    def histogram(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ],
        column: Annotated[
            str, Field(description='Column name from SQL result to use for the histogram')
        ],
        color: Annotated[
            str | None, Field(description='Optional; column name from SQL result to use for drawing multiple colored histograms representing another dimension')
        ] = None,
        nbins: Annotated[
            int | None, Field(description='Optional; number of bins')
        ] = None,
    ) -> str:
        """
        Run query against specified source and make a histogram using result
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name in the SQL query for csv sources.
        Use 'PARQUET' as the table name in the SQL query for parquet sources.
    
        This will return an image of the plot
        """

        try:
            df = self._get_df_from_source(source_id, query)
            fig = px.histogram(df, x=column, color=color, nbins=nbins)
            fig.update_xaxes(autotickangles=[0, 45, 60, 90])

            return _fig_to_image(fig)
        except Exception as e:
            return str(e)


    def strip_plot(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ],
        x: Annotated[
            str, Field(description='Column name from SQL result to use for x-axis')
        ],
        y: Annotated[
            str, Field(description='Column name from SQL result to use for y-axis')
        ],
        color: Annotated[
            str | None, Field(description='Optional column name from SQL result to show multiple colored strips representing another dimension')
        ] = None,
    ) -> str:
        """
        Run query against specified source and make a strip plot using result
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name in the SQL query for csv sources.
        Use 'PARQUET' as the table name in the SQL query for parquet sources.
    
        This will return an image of the plot
        """

        try:
            df = self._get_df_from_source(source_id, query)
            fig = px.strip(df, x=x, y=y, color=color)
            fig.update_xaxes(autotickangles=[0, 45, 60, 90])

            return _fig_to_image(fig)
        except Exception as e:
            return str(e)


    def box_plot(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ],
        x: Annotated[
            str, Field(description='Column name from SQL result to use for x-axis')
        ],
        y: Annotated[
            str, Field(description='Column name from SQL result to use for y-axis')
        ],
        color: Annotated[
            str | None, Field(description='Optional column name from SQL result to show multiple colored bars representing another dimension')
        ] = None,
    ) -> str:
        """
        Run query against specified source and make a box plot using result
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name in the SQL query for csv sources.
        Use 'PARQUET' as the table name in the SQL query for parquet sources.
    
        This will return an image of the plot
        """

        try:
            df = self._get_df_from_source(source_id, query)
            fig = px.box(df, x=x, y=y, color=color)
            fig.update_xaxes(autotickangles=[0, 45, 60, 90])

            return _fig_to_image(fig)
        except Exception as e:
            return str(e)


    def bar_plot(self,
        source_id: Annotated[
            str, Field(description='The data source to run the query on')
        ],  
        query: Annotated[
            str, Field(description='SQL query to run on the data source')
        ],
        x: Annotated[
            str, Field(description='Column name from SQL result to use for x-axis')
        ],
        y: Annotated[
            str, Field(description='Column name from SQL result to use for y-axis')
        ],
        color: Annotated[
            str | None, Field(description='Optional column name from SQL result to use as a 3rd dimension by splitting each bar into colored sections')
        ] = None,
        orientation: Annotated[
            str, Field(description="Orientation of the box plot, use 'v' for vertical (default) and 'h' for horizontal. Be mindful of choosing the correct X and Y columns as per orientation")
        ] = 'v',
    ) -> str:
        """
        Run query against specified source and make a bar plot using result
        For both csv and parquet sources, use DuckDB SQL syntax
        Use 'CSV' as the table name in the SQL query for csv sources.
        Use 'PARQUET' as the table name in the SQL query for parquet sources.
    
        This will return an image of the plot
        """

        try:
            df = self._get_df_from_source(source_id, query)
            fig = px.bar(df, x=x, y=y, color=color, orientation=orientation)
            fig.update_xaxes(autotickangles=[0, 45, 60, 90])

            return _fig_to_image(fig)
        except Exception as e:
            return str(e)


if __name__=="__main__":
    print(ImageContent)
